# Generated by Django 2.2.14 on 2020-08-28 14:34

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Lock',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('lock_name', models.CharField(max_length=128, unique=True, verbose_name='Lock Name')),
                ('worker_name', models.CharField(max_length=128, verbose_name='Worker Name')),
                ('lock_time', models.DateTimeField(verbose_name='Lock Time')),
                ('expire_time', models.DateTimeField(verbose_name='Expired Time')),
            ],
            options={
                'verbose_name': 'Lock',
                'verbose_name_plural': 'Locks',
            },
        ),
    ]



import time
import json
from django.test import TestCase
from django.test import Client
from django.urls import reverse_lazy as reverse
from django_db_lock.server import django_db_lock_default_server
from .models import Lock

class TestDjangoDbLock(TestCase):

    def setUp(self):
        self.url_acquire_lock = reverse("django_db_lock.acquireLock")
        self.url_release_lock = reverse("django_db_lock.releaseLock")
        self.url_get_lock_info = reverse("django_db_lock.getLockInfo")
        self.url_clear_expired_locks = reverse("django_db_lock.clearExpiredLocks")
        self.browser = Client()

    def test01(self):
        assert django_db_lock_default_server.acquire_lock("lock01", "worker01", 100)
        assert django_db_lock_default_server.acquire_lock("lock01", "worker01", 100) == False
        assert django_db_lock_default_server.release_lock("lock01", "worker01")

    def test02(self):
        assert django_db_lock_default_server.acquire_lock("lock01", "worker01", 1)
        time.sleep(2)
        assert django_db_lock_default_server.acquire_lock("lock01", "worker01", 1)

    def test03(self):
        time.sleep(2)
        assert django_db_lock_default_server.acquire_lock("lock01", "worker01", 100)
        assert django_db_lock_default_server.release_lock("lock01", "worker01")
        assert django_db_lock_default_server.release_lock("lock01", "worker01")

    def test04(self):
        django_db_lock_default_server.acquire_lock("lock01", "worker01", 100)
        info = django_db_lock_default_server.get_lock_info("lock01")
        assert info["lockName"] == "lock01"
        assert info["workerName"] == "worker01"

    def test05(self):
        time.sleep(2)
        
        browser = Client()
        response = browser.get(self.url_acquire_lock, {"lockName": "lock01", "workerName": "worker01", "timeout": 300})
        data = json.loads(response.content)
        assert data["result"]

        browser = Client()
        response = browser.get(self.url_acquire_lock, {"lockName": "lock01", "workerName": "worker01", "timeout": 300})
        data = json.loads(response.content)
        assert data["result"] is False

        browser = Client()
        response = browser.get(self.url_get_lock_info, {"lockName": "lock01"})
        data = json.loads(response.content)
        assert data["result"]
        assert data["result"]["lockName"] == "lock01"

        browser = Client()
        response = browser.get(self.url_release_lock, {"lockName": "lock01", "workerName": "worker01"})
        data = json.loads(response.content)
        assert data["result"]

    def test09(self):
        browser = Client()
        response = browser.get(self.url_clear_expired_locks)
        data = json.loads(response.content)
        assert data["result"]



from django.conf import settings
from django.contrib import admin
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.contrib.humanize.templatetags import humanize

from .models import Lock
from . import settings

class LockAdmin(admin.ModelAdmin):
    list_display = ["lock_name", "worker_name", "lock_time", "expire_time", "time_remaining"]
    search_fields = ["lock_name", "worker_name"]

    def time_remaining(self, obj):
        expire_time = timezone.make_naive(obj.expire_time)
        return humanize.naturaltime(expire_time)
    time_remaining.short_description = _("Time Remaining")
    time_remaining.admin_order_field = "expire_time"

if settings.DJANGO_DB_LOCK_AUTO_REGISTER_ADMIN and settings.DJANGO_DB_LOCK_AUTO_REGISTER_MODEL:
    admin.site.register(Lock, LockAdmin)



from . import settings

if settings.DJANGO_DB_LOCK_AUTO_REGISTER_SERVICES and settings.DJANGO_DB_LOCK_AUTO_REGISTER_MODEL:
    from .server import django_db_lock_default_server
    urlpatterns = django_db_lock_default_server.get_urls()
else:
    urlpatterns = []



from django.conf import settings

DJANGO_DB_LOCK_AUTO_REGISTER_MODEL = getattr(settings, "DJANGO_DB_LOCK_AUTO_REGISTER_MODEL", True)
DJANGO_DB_LOCK_AUTO_REGISTER_ADMIN = getattr(settings, "DJANGO_DB_LOCK_AUTO_REGISTER_ADMIN", True)
DJANGO_DB_LOCK_AUTO_REGISTER_SERVICES = getattr(settings, "DJANGO_DB_LOCK_AUTO_REGISTER_SERVICES", True)
DJANGO_DB_LOCK_APP_LABEL = getattr(settings, "DJANGO_DB_LOCK_APP_LABEL", "django_db_lock")
DJANGO_DB_LOCK_API_SERVER = getattr(settings, "DJANGO_DB_LOCK_API_SERVER", None)
DJANGO_DB_LOCK_ACQUIRE_LOCK_PATH = getattr(settings, "DJANGO_DB_LOCK_ACQUIRE_LOCK_PATH", "acquireLock")
DJANGO_DB_LOCK_RELEASE_LOCK_PATH = getattr(settings, "DJANGO_DB_LOCK_RELEASE_LOCK_PATH", "releaseLock")

DJANGO_DB_LOCK_REDIS_CACHE_NAME = getattr(settings, "DJANGO_DB_LOCK_REDIS_CACHE_NAME", "default")
DJANGO_DB_LOCK_USA_REDIS_CACHE = getattr(settings, "DJANGO_DB_LOCK_USA_REDIS_CACHE", True)
DJANGO_DB_LOCK_REDIS_CACHE_PREFIX = getattr(settings, "DJANGO_DB_LOCK_REDIS_CACHE_PREFIX", "django-db-lock:locks")




import time
import logging
import datetime

from django.urls import path
from django.utils import timezone
from django.db.utils import IntegrityError
from django.db import transaction

try:
    from django_redis import get_redis_connection
except ImportError:
    get_redis_connection = None

from django_apiview.views import apiview

from . import settings

logger = logging.getLogger(__name__)

class DjangoDbLockServer(object):

    def __init__(self, model, use_redis_cache=None, redis_cache_name=None, redis_cache_prefix=None):
        self.model = model
        self.use_redis_cache = use_redis_cache or settings.DJANGO_DB_LOCK_USA_REDIS_CACHE
        self.redis_cache_name = redis_cache_name or settings.DJANGO_DB_LOCK_REDIS_CACHE_NAME
        self.redis_cache_prefix = redis_cache_prefix or settings.DJANGO_DB_LOCK_REDIS_CACHE_PREFIX
        self.redis_cache_enabled = self.get_redis_cache_enabled()

    def get_redis_cache_enabled(self):
        if not self.use_redis_cache:
            return False
        if not get_redis_connection:
            return False
        try:
            db = get_redis_connection(self.redis_cache_name)
            return db.ping()
        except Exception:
            return False

    def get_redis_connection(self):
        if self.enable_redis_cache:
            return get_redis_connection(self.redis_cache_name)
        else:
            return None

    def get_redis_lock_cache_key(self, lock_name):
        return ":".join([self.redis_cache_prefix, lock_name])
    
    def clear_expired_locks(self):
        try:
            return self._clear_expired_locks()
        except Exception as error:
            logger.exception("clear_expired_locks failed: error_message={}".format(error))

    def _clear_expired_locks(self):
        now_time = timezone.now()
        with transaction.atomic():
            self.model.objects.filter(expire_time__lte=now_time).delete()


    def acquire_lock(self, lock_name, worker_name, timeout):
        self.clear_expired_locks()
        expire_at = int(time.time() + timeout)
        result1 = self.acquire_lock_in_redis(lock_name, worker_name, expire_at)
        if not result1:
            return result1
        result2 = self.acquire_lock_in_database(lock_name, worker_name, expire_at)
        return result2

    def acquire_lock_in_redis(self, lock_name, worker_name, expire_at):
        try:
            return self._acquire_lock_in_redis(lock_name, worker_name, expire_at)
        except Exception as error:
            logger.exception("acquire_lock_in_redis failed, lock_name={} worker_name={} expire_at={} error_message={0}".format(lock_name, worker_name, expire_at, error))
            return False

    def _acquire_lock_in_redis(self, lock_name, worker_name, expire_at):
        if not self.redis_cache_enabled:
            return True
        db = self.get_redis_connection()
        key = self.get_redis_lock_cache_key(lock_name)
        flag = db.setnx(key, worker_name)
        if flag:
            db.expireat(key, expire_at)
        return flag

    def acquire_lock_in_database(self, lock_name, worker_name, expire_at):
        try:
            return self._acquire_lock_in_database(lock_name, worker_name, expire_at)
        except Exception as error:
            logger.exception("acquire_lock_in_database failed, lock_name={} worker_name={} error_message={0}".format(lock_name, worker_name, error))
            return False

    def _acquire_lock_in_database(self, lock_name, worker_name, expire_at): 
        lock = self.model()
        lock.lock_name = lock_name
        lock.worker_name = worker_name
        lock.lock_time = timezone.now()
        lock.expire_time = timezone.make_aware(datetime.datetime.fromtimestamp(expire_at))
        try:
            with transaction.atomic():
                lock.save()
            return True
        except IntegrityError:
            return False

    def release_lock(self, lock_name, worker_name):
        self.clear_expired_locks()
        result1 = self.release_lock_in_database(lock_name, worker_name)
        result2 = self.release_lock_in_redis(lock_name, worker_name)
        return result1 and result2

    def release_lock_in_redis(self, lock_name, worker_name):
        try:
            return self._release_lock_in_redis(lock_name, worker_name)
        except Exception:
            logger.exception("release_lock_in_redis failed: lock_name={} worker_name={} error_message={}".format(lock_name, worker_name, error))
            return False

    def _release_lock_in_redis(self, lock_name, worker_name):
        if self.redis_cache_enabled:
            db = self.get_redis_connection()
            key = self.get_redis_lock_cache_key(lock_name)
            real_worker_name = db.get(key)
            if real_worker_name == worker_name:
                db.delete(key)
        return True

    def release_lock_in_database(self, lock_name, worker_name):
        try:
            return self._release_lock_in_database(lock_name, worker_name)
        except Exception as error:
            logger.exception("release_lock_in_database failed: lock_name={} worker_name={} error_message={}".format(lock_name, worker_name, error))
            return False

    def _release_lock_in_database(self, lock_name, worker_name):
        try:
            lock = self.model.objects.get(lock_name=lock_name, worker_name=worker_name)
            with transaction.atomic():
                lock.delete()
        except self.model.DoesNotExist:
            pass
        return True

    def get_lock_info(self, lock_name):
        try:
            lock = self.model.objects.get(lock_name=lock_name)
            return {
                "pk": lock.pk,
                "lockName": lock.lock_name,
                "workerName": lock.worker_name,
                "lockTime": lock.lock_time,
                "expireTime": lock.expire_time,
            }
        except self.model.DoesNotExist:
            return None

    def get_urls(self, prefix="django_db_lock"):
        return [
            path('acquireLock', self.getAcquireLockView(), name=prefix + ".acquireLock"),
            path('releaseLock', self.getReleaseLockView(), name=prefix + ".releaseLock"),
            path('getLockInfo', self.getGetLockInfoView(), name=prefix + ".getLockInfo"),
            path('clearExpiredLocks', self.GetClearExpiredLocksView(), name=prefix + ".clearExpiredLocks"),
        ]

    def getAcquireLockView(self):
        @apiview
        def acquireLock(lockName, workerName, timeout:int):
            result = self.acquire_lock(lockName, workerName, timeout)
            return result
        return acquireLock

    def getReleaseLockView(self):
        @apiview
        def releaseLock(lockName, workerName):
            result = self.release_lock(lockName, workerName)
            return result
        return releaseLock

    def getGetLockInfoView(self):
        @apiview
        def getLockInfo(lockName):
            info = self.get_lock_info(lockName)
            return info
        return getLockInfo

    def GetClearExpiredLocksView(self):
        @apiview
        def clearExpiredLocks():
            self.clear_expired_locks()
            return True
        return clearExpiredLocks

if settings.DJANGO_DB_LOCK_AUTO_REGISTER_SERVICES and settings.DJANGO_DB_LOCK_AUTO_REGISTER_MODEL:
    from .models import Lock
    django_db_lock_default_server = DjangoDbLockServer(Lock)
